import sqlite3
import os
import pandas as pd
import numpy as np
from pandas.core.frame import DataFrame
from sensiml.dclproj.datasegments import DataSegments, DataSegment
import uuid
from collections import OrderedDict
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
from typing import List, Optional, Dict, Tuple


class DCLProject:
    def __init__(self):
        self._path = None
        self._conn = None
        self._tables = None
        self._verbose = False

    @property
    def verbose(self):
        return self._verbose

    @verbose.setter
    def verbose(self, value: bool):
        if isinstance(value, bool):
            self._verbose = value
        else:
            print("verbose must be either True or False.")

    @property
    def data_dir(self) -> str:

        if self._path is None:
            raise Exception("Path is not set!")

        return os.path.join(self._path, "data")

    def _set_table_list(self):
        cursorObj = self._conn.cursor()

        cursorObj.execute('SELECT name from sqlite_master where type= "table"')

        self._tables = [x[0] for x in cursorObj.fetchall()]

    def create_connection(self, db_file: str) -> None:
        """create a database connection to the SQLite database
            specified by db_file
        :param db_file: database file
        :return: Connection object or None
        """
        self._path = os.path.dirname(db_file)
        if not os.path.exists(db_file):
            print("database file not found at {}".format(db_file))
        conn = None
        try:
            conn = sqlite3.connect(db_file)
        except Exception as e:
            print(e)
            raise e

        self._conn = conn
        self._set_table_list()

    def _execute_query(self, query: str, header: str = None) -> DataFrame:

        cur = self._conn.cursor()

        if self.verbose:
            print(query)

        cur.execute(query)

        if header is None:
            header = [x[0] for x in cur.description]

        rows = cur.fetchall()

        results = []
        for row in rows:
            results.append(row)

        df = pd.DataFrame(results, columns=header)

        if "uuid" in header:
            df["uuid"] = df["uuid"].apply(lambda x: uuid.UUID(bytes=x) if x else None)

        if "local_status" in header:
            local_status = {
                0: "To Add",
                1: "To Update",
                "Synced": "Synced",
                2: "To Delete",
                None: "Synced",
            }
            df["local_status"] = (
                df["local_status"].fillna("Synced").apply(lambda x: local_status[x])
            )

        return df

    def _list_table_raw(self, tablename: str) -> DataFrame:

        if tablename not in self._tables:
            print("Table is not part of the database.")
            return None

        query = "SELECT * FROM {tablename} ".format(tablename=tablename)

        return self._execute_query(query)

    def _list_table(
        self,
        tablename: str,
        fields: List[str],
        fk_fields: Optional[Dict] = None,
        query_filter: Optional[str] = None,
    ):

        if tablename not in self._tables:
            print("Table is not part of the database.")
            return None

        select_fields = ", ".join(
            ["{}.{}".format(tablename, field) for field in fields]
        )

        if fk_fields:
            select_fields += ", " + ", ".join(
                [
                    "{}.name".format(fk_table)
                    if fk_table != "LabelValue"
                    else "{}.value".format(fk_table)
                    for fk_table in fk_fields.keys()
                ]
            )

        query_select = "SELECT {select_fields} FROM {tablename} ".format(
            select_fields=select_fields, tablename=tablename
        )

        query_join = ""

        if fk_fields:
            join_fields = []
            for fk_table, fk_field in fk_fields.items():
                join_fields.append(
                    "JOIN {fk_table} ON {fk_table}.id = {tablename}.{fk_field} ".format(
                        fk_table=fk_table, tablename=tablename, fk_field=fk_field
                    )
                )
            query_join += " ".join(join_fields)

        header = None

        if query_filter:
            query_where = (
                query_filter  # + "AND {}.local_status != 2 ".format(tablename)
            )
        else:
            query_where = ""  # "WHERE {}.local_status != 2 ".format(tablename)

        if fk_fields:
            header = fields + list(fk_fields.values())

        return self._execute_query(query_select + query_join + query_where, header)

    def list_captures(self) -> DataFrame:
        fields = [
            "uuid",
            "name",
            "file_size",
            "number_samples",
            "set_sample_rate",
            "created_at",
            "local_status",
            "last_modified",
        ]

        fk_fields = OrderedDict([("CaptureConfiguration", "capture_configuration")])

        return self._list_table("Capture", fields=fields, fk_fields=fk_fields)

    def list_sessions(self) -> DataFrame:
        fields = [
            "id",
            "name",
            "parameters",
            "custom",
            "preprocess",
            "created_at",
            "local_status",
            "last_modified",
        ]

        fk_fields = OrderedDict([])

        return self._list_table("Segmenter", fields, fk_fields)

    def list_capture_segments(
        self, capture: Optional[str] = None, session: Optional[str] = None
    ) -> DataFrame:

        query_filter = ""

        if capture and session:
            query_filter = 'WHERE Capture.name="{}" '.format(
                capture
            ) + 'AND Segmenter.name = "{0}" '.format(session)

        elif capture:
            query_filter = 'WHERE Capture.name="{}" '.format(capture)
        elif session:
            query_filter += 'WHERE Segmenter.name = "{0}" '.format(session)

        fields = [
            "uuid",
            "capture_sample_sequence_start",
            "capture_sample_sequence_end",
            "local_status",
            "created_at",
            "last_modified",
        ]

        fk_fields = OrderedDict(
            [
                ("Segmenter", "segmenter"),
                ("Capture", "capture"),
                ("LabelValue", "label_value"),
            ]
        )

        return self._list_table("CaptureLabelValue", fields, fk_fields, query_filter)

    def get_capture_metadata(self, capture: str) -> DataFrame:
        "returns a dataframe containing the metadata associated with this capture"

        fields = ["uuid"]

        fk_fields = OrderedDict(
            [("Capture", "capture"), ("LabelValue", "label_value"), ("Label", "label")]
        )

        df = self._list_table("CaptureMetadataValue", fields, fk_fields=fk_fields)

        return df[df["capture"] == capture]

    def get_capture_segments(self, capture: str, session: str):
        """
        Returns a DataSegment object of the specified capture and session


        Args:
            capture_name (str): name of capture
            session (str):  name of session where the labels are
            recognized (Dataframe): results from recognize signal on this capture
            class_map (dict): The class map describing how to map the labels to integer values
            groups (dict): a dict of key value pairs where the key is the original label the value is the label it was should be renamed to

        """

        labels = self.list_capture_segments(
            capture=capture, session=session
        ).sort_values(by="capture_sample_sequence_start")

        # TODO: read wave file if file is .wav
        tmp_df = pd.read_csv(os.path.join(self.data_dir, capture), index_col="sequence")

        M = DataSegments()

        for index, label in enumerate(
            labels[
                [
                    "capture_sample_sequence_start",
                    "capture_sample_sequence_end",
                    "uuid",
                    "label_value",
                    "capture",
                ]
            ].values
        ):
            M[index] = DataSegment(
                tmp_df.loc[label[0] : label[1]],
                segment_index=index,
                label=label[3],
                uuid=label[2],
                capture=label[4],
            )

        return M

    def get_capture(self, capture_name: str) -> DataFrame:
        """
        Returns the capture as a dataframe

        Args:
            capture_name (str): name of capture

        """

        # TODO: read wave file if file is .wav
        tmp_df = pd.read_csv(
            os.path.join(self.data_dir, capture_name), index_col="sequence"
        )

        return tmp_df

    def get_captures(self, capture_names: Optional[List[str]] = None) -> DataFrame:
        """
        Returns the capture as a single dataframe

        Args:
            capture_names (List[str]): name of captures to return, if None returns all captures

        """
        if capture_names is None:
            capture_names = os.listdir(self.datadir)

        return pd.concat(
            map(
                pd.read_csv,
                [os.path.join(self.datadir, capture) for capture in capture_names],
            )
        ).reset_index(drop=True)

    def plot_segment_labels(
        self,
        capture_names: List[str],
        session: str,
        columns: Optional[List[str]] = None,
        transforms: Optional[List[Dict]] = None,
        figsize: Tuple = (30, 8),
        ylim: Optional[Tuple] = None,
    ) -> None:
        """
        Creates a plot of the labels and raw signal data for a session and one or more capture files.

        Args:
            capture_names (List[str]): List if captures to plot
            session (str):  name of session to pull the labels from
            columns (list): a list of columns to plot from the data
             label the value is the label it was should be renamed to
            transforms: additional fucntions to apply prior to plotting the data


        example of a single transform::

        ```
            def sum_columns(x):
                return x[['Column1', 'Column2']].sum()

            func = {'name':'sum','func':sum_columns, 'type':'transform'}

            dcl.plot_segment_labels(captures, session_name, columns=['sum], transforms=[func])
        ```

        example of multiple transforms
        transforms will be executed in order

        ```

            def sum_columns(x):
                return x[['Column1', 'Column2']].sum()

            def remove_offset(x):
                x['Column1']-=100
                x['Column2']-=200
                return x

            func2 = {'name':'filter','func':remove_offset, "type":'filter', "columns":['Column1', 'Column2']}
            func1 = {'name':'sum','func':sum_columns, "type":'transform'}


            dcl.plot_segment_labels(captures, session_name, columns=['Column1', 'Column2', 'sum'], transforms=[func1, func2])


        """

        sessions = self.list_sessions()
        session_id = sessions[sessions.name == session].id.values[0]
        clv = self._list_table_raw("CaptureLabelValue")
        label_value_ids = clv[clv.segmenter == session_id].label_value.unique()
        lv = self._list_table_raw("LabelValue")
        labels = sorted(lv[lv.id.isin(label_value_ids)].value.values.tolist())
        class_map = {label: index for index, label in enumerate(labels)}

        M = []
        current_start = 0

        plt.figure(figsize=figsize)
        currentAxis = plt.gca()

        if isinstance(capture_names, str):
            capture_names = [capture_names]

        for capture_name in capture_names:
            segments = self.list_capture_segments(capture_name, session)
            capture = self.get_capture(capture_name)
            M.append(capture)

            capture.describe().loc["max"].max()

            cmap = plt.cm.rainbow

            labels = list(class_map.keys())

            delta = 1 / len(labels)
            label_float = np.arange(0, 1, delta)
            label_float[-1] = 1.0

            label_colors = {
                labels[index]: cmap(x) for index, x in enumerate(label_float)
            }
            label_legend = [
                Line2D([0], [0], color=cmap(x), lw=4) for x in label_float
            ] + [Line2D([0], [0], color="white", lw=4)]

            for _, seg in segments.iterrows():
                y_origin = (
                    capture.describe().loc["min"].min() if ylim is None else ylim[0]
                )
                x_origin = seg["capture_sample_sequence_start"] + current_start
                x_final = (
                    seg["capture_sample_sequence_end"]
                    - seg["capture_sample_sequence_start"]
                )
                y_final = (
                    capture.describe().loc["max"].max() - y_origin
                    if ylim is None
                    else ylim[1] - ylim[0]
                )

                currentAxis.add_artist(
                    plt.Rectangle(
                        (x_origin, y_origin),
                        x_final,
                        y_final,
                        alpha=0.2,
                        color=label_colors[seg["label_value"]],
                    )
                )

            currentAxis.legend(label_legend, labels + [""])

            currentAxis.text(
                current_start,
                0 if ylim is None else ylim[0],
                capture_name,
                style="italic",
                rotation=90,
                # bbox={"facecolor": "grey", "alpha": 0.5, "pad": 10},
            )

            current_start += capture.shape[0]

        capture = pd.concat(M)

        if transforms is not None:
            for func in transforms:
                if func["type"] == "transform":
                    capture[func["name"]] = capture.apply(
                        lambda x: func["func"](x), axis=func.get("axis", 1)
                    )

                if func["type"] == "filter":
                    capture[func["columns"]] = capture.apply(
                        lambda x: func["func"](x), axis=func.get("axis", 1)
                    )

        if columns is None:
            columns = capture.columns

        capture[columns].reset_index(drop=True).plot(
            figsize=figsize,
            ax=currentAxis,
            title="session: '{session}'".format(capture=capture_name, session=session),
            ylim=ylim,
        )

        return capture.reset_index(drop=True)



from sensiml.widgets.auto_sense import AutoSenseWidget
from sensiml.widgets.download_widget import DownloadWidget
from sensiml.widgets.download_widget_2 import DownloadWidgetVersion2
from sensiml.widgets.query_widget import QueryWidget
from sensiml.widgets.flash_widget import FlashWidget
from sensiml.widgets.model_visualization_widget import ModelVisualizationWidget
from sensiml.widgets.model_explore_widget import ModelExploreWidget
from sensiml.widgets.dashboard_widget import DashBoard
from sensiml.widgets.pipeline_builder_widget import PipelineBuilderWidget
from sensiml.widgets.hierarchical_model_visualization import (
    HierarchicalModelVisualizationWidget,
)


__all__ = [
    "AutoSenseWidget",
    "DownloadWidget",
    "QueryWidget",
    "FlashWidget",
    "ModelVisualizationWidget",
    "DashBoard",
    "ModelExploreWidget",
    "PipelineBuilderWidget",
    "HierarchicalModelVisualizationWidget",
    "DownloadWidgetVersion2",
]



import logging
import os
import re
from tempfile import NamedTemporaryFile
from appdirs import user_config_dir
import json

try:
    from qgrid import show_grid
except:
    pass

import pandas as pd
from pandas import DataFrame

import sensiml.base.utility as utility
from sensiml.datamanager import (
    Functions,
    PlatformDescriptions,
    ClientPlatformDescriptions,
    PipelineSeeds,
    Projects,
    SegmenterSet,
)
from sensiml.datamanager.knowledgepack import (
    get_knowledgepack,
    get_knowledgepacks,
    delete_knowledgepack,
)
from sensiml.datasets import DataSets
from sensiml.connection import Connection

# from sensiml.base.exceptions import *
from sensiml.pipeline import Pipeline
from sensiml.base.snippets import Snippets, function_help
from sensiml.datamanager.captures import CaptureExistsError
from time import sleep

config_dir = user_config_dir(__name__.split(".")[0], False)

logger = logging.getLogger("SensiML")


def print_list(func):
    """This is a wrapper for printing out lists of objects stored in SensiML Cloud"""

    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if len(result.keys()) == 0 and kwargs.get("silent", False) == False:
            print(
                "No {} stored on SensiML Cloud for this project.".format(
                    " ".join(func.__name__.split("_")[1:])
                )
            )
            return None
        if kwargs.get("get_objects", False) is True:
            return result

        tmp_dict = [
            {
                "Name": name,
                "UUID": item.uuid,
                "Created": item.created_at,
                "Last Modified": item.created_at
                if not hasattr(item, "last_modified")
                else item.last_modified,
            }
            for name, item in result.items()
        ]

        extra = {}
        additional_keys = []
        if func.__name__ == "list_queries":
            extra = [
                {"Cached": True if item.cache else ""} for name, item in result.items()
            ]
            additional_keys = ["Cached"]

        for index, value in enumerate(extra):
            tmp_dict[index].update(value)

        return DataFrame(
            tmp_dict,
            columns=["Name", "Last Modified", "Created", "UUID"] + additional_keys,
        )

    return wrapper


class SensiML(object):
    """Entrance to SensiML Analytic Suite"""

    # Server Minimum Version For Campatibility
    __minversion__ = "2019.1.0"

    def __init__(
        self,
        server="https://sensiml.cloud/",
        path="connect.cfg",
        use_jedi=False,
        **kwargs
    ):

        self._project = None
        self._pipeline = None
        auth_url = server + "oauth/"

        self._connection = Connection(
            server=server, auth_url=auth_url, path=path, **kwargs
        )

        self.validate_client_version(kwargs.get("skip_validate", True))
        self.projects = Projects(self._connection)
        self.datasets = DataSets()
        self.functions = Functions(self._connection)
        self.platforms = PlatformDescriptions(self._connection)
        self.platforms_v2 = ClientPlatformDescriptions(self._connection)
        self.seeds = PipelineSeeds(self._connection)
        self.snippets = Snippets(
            self.list_functions(kp_functions=False, qgrid=False),
            self.functions.function_list,
            self.seeds,
        )
        self._feature_files = None

        if use_jedi is False:
            self.setup_jedi_false()

    def setup_jedi_false(self):
        """This is a temporary bug fix in ipython autocomplete"""
        try:
            mgc = get_ipython().magic
            mgc("%config Completer.use_jedi = False")
        except:
            pass

    def validate_client_version(self, skip_validate=True):
        """Perform a Validation check to see if this version of SensiML is up to date with the latest."""

        if skip_validate:
            return

        url = "version"

        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)

        if response_data["SensiML_Python_Library_Windows"] != self.__minversion__:
            print(
                "SensiML Python Client Library is out of date. Update before connecting to the server."
            )
            print("In order to update sensiml client use 'pip install sensiml -U'.")
            print(
                "For Instrunctions on updating SensiML Client for the Analytic Studio visit https://sensiml.atlassian.net/wiki/spaces/SS/pages/115179614/Updating+SensiML+Client"
            )
            raise Exception("Python Client is out of date.")

    def logout(self, name=None):
        """Logs out of the current connection."""
        if name is None:
            name = self._connection.server_name

        Connection.logout(name)

    def get_function(self, name):
        """Gets a function method call"""
        return self.functions.function_list[name]

    def function_description(self, name):
        """Gets a description of a pipeline function."""
        print(self.functions.create_function_call(name).__doc__)

    def function_help(self, name):
        """Prints a shortened description of a function."""
        print(function_help(self.functions.function_list[name]))

    def list_functions(
        self, functype=None, subtype=None, kp_functions=False, qgrid=False
    ):
        """Lists all of the functions available on SensiML Cloud

        Returns:
            Dataframe

        Args:
            functype (str, None): Return only functions with the specified type. ie. "Segmenter"
            subtype (str, None): Return only functions with the specified subtype. ie. "Sensor"
            kp_functions (boolean, True): Return only functions that run on tbe loaded to a device.
            Excludes functions such as feature selection and model training.
        """

        df = (
            DataFrame(
                [
                    {
                        "NAME": f.name,
                        "TYPE": f.type,
                        "DESCRIPTION": f.description.lstrip("\n").lstrip(" ")
                        if f.description
                        else "",
                        "SUBTYPE": f.subtype,
                        "KP FUNCTION": f.has_c_version,
                        "UUID": f.uuid,
                        "LIBRARY": f.library_pack,
                    }
                    for name, f in self.functions.function_list.items()
                ]
            )
            .sort_values(by=["TYPE", "SUBTYPE"])
            .reset_index(drop=True)[
                [
                    "NAME",
                    "TYPE",
                    "SUBTYPE",
                    "DESCRIPTION",
                    "KP FUNCTION",
                    "UUID",
                    "LIBRARY",
                ]
            ]
        )

        if functype:
            df = df[df["TYPE"] == functype]

        if subtype:
            df = df[df["SUBTYPE"] == subtype]

        if kp_functions:
            df = df[df["KP FUNCTION"] == True][
                ["NAME", "TYPE", "SUBTYPE", "DESCRIPTION", "LIBRARY"]
            ]

        if qgrid:
            return show_grid(df.reset_index(drop=True))
        else:
            return df.reset_index(drop=True)

    def delete_project(self):
        """Deletes a project"""
        if self._project is not None:
            self._project.delete()

    @print_list
    def list_projects(self, get_objects=False, silent=False):
        """Lists all of the projects on SensiML Cloud

        Returns:
            DataFrame: projects on SensiML Cloud
        """
        return self.projects.build_project_dict()

    def list_segmenters(
        self,
    ):
        if self._project is None:
            print("project must be set to list segmenters.")
            return None

        segmenters = SegmenterSet(self._connection, self._project)

        if not len(segmenters.objs):
            print("No segmenters stored on the Cloud.")
            return None

        return segmenters.to_dataframe()

    def list_seeds(self):
        """Lists all of the pipeline seeds on SensiML Cloud

        Returns:
            DataFrame: pipeline seeds on SensiML Cloud
        """
        return DataFrame(
            {
                "Name": [s.name for s in self.seeds],
                "Description": [s.description for s in self.seeds],
            }
        )

    def get_automl_parameter_inventory(self):
        """
        inventory paramenters for AutoML

        Returns:
            DataFrame: Automl parameter inventory
        """

        url = "automl/inventory/"

        response = self._connection.request("get", url)

        response_data, err = utility.check_server_response(response)

        return DataFrame(response_data)

    def get_knowledgepack(self, uuid):
        """retrieve knowledgepack by uuid from the server associated with current project

        Args:
            uuid (str): unique identifier for knowledgepack

        Returns:
            TYPE: a knowledgepack object
        """

        return get_knowledgepack(uuid, self._connection)

    def delete_knowledgepack(self, uuid):
        """delete knowledgepack by uuid from the server associated with current project

        Args:
            uuid (str): unique identifier for knowledgepack

        Returns:
            TYPE: a knowledgepack object
        """

        return delete_knowledgepack(uuid, self._connection)

    @property
    def project(self):
        """The active project"""
        return self._project

    def get_featurefile(self, uuid):
        """Get a featruefile by uuid

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.get_featurefile(uuid)

    def get_datafile(self, uuid):
        """Get a datafile by uuid

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.get_featurefile(uuid)

    @print_list
    def list_featurefiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._feature_files.build_featurefile_list()

    @print_list
    def list_datafiles(self, get_objects=False, silent=False):
        """List all feature and data files for the active project.

        Args:
            get_objects (bool, False): Also return the featurefile objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._feature_files.build_datafile_list()

    @print_list
    def list_captures(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._captures.build_capture_list()

    @print_list
    def list_capture_configurations(self, get_objects=False):
        """List all captures for the active project

        Args:
            get_objects (bool, False): Also return the capture objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")

        return self._project._capture_configurations.build_capture_list()

    @print_list
    def list_sandboxes(self, get_objects=False):
        """List all sandboxes for the active project.

        Args:
            get_objects (bool, False): Also return the sandbox objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._sandboxes.build_sandbox_list()

    @print_list
    def list_queries(self, get_objects=False):
        """List all queries for the active project.

        Args:
            get_objects (bool, False): Also return the query objects.

        """
        if self._project is None:
            raise Exception("Project must be set to perform this action.")
        return self._project._queries.build_query_list()

    @project.setter
    def project(self, name):
        self._project = self.projects.get_or_create_project(name)

    @property
    def pipeline(self):
        """The active pipeline"""
        return self._pipeline

    @pipeline.setter
    def pipeline(self, name):
        if self._project is None:
            raise Exception("Project must be set before a pipeline can be created")

        self._pipeline = Pipeline(self, name=name)

    def create_query(
        self,
        name,
        columns=[],
        metadata_columns=[],
        metadata_filter="",
        segmenter=None,
        label_column="",
        combine_labels=None,
        force=False,
        renderer=None,
        capture_configurations="",
    ):
        """Create a query to use as input data in a pipeline.

        Args:
            name (str): Name of the query.
            columns (list, optional): Columns to add to the query result.
            metadata_columns (list, optional): Metadata to add to the query result.
            metadata_filter (str, optional): Filter to apply to the query.
            segmenter (int, optional): Segmenter to filter query by.
            force (bool, False): If True overwrite the query on kb cloud.

        Returns:
            object: Returns a query object that was created.
        """

        query = self.project.queries.get_query_by_name(name)

        new = False
        if query is not None and not force:
            raise QueryExistsException(
                "Query already exists. Set force=True to overwrite."
            )
        elif query is not None and force:
            query.columns.clear()
            query.metadata_columns.clear()
            query.metadata_filter = ""
            query.label_column = ""
            query.capture_configurations = ""
            query.segmenter = None

        else:
            query = self.project.queries.new_query()
            query.name = name
            new = True

        for col in columns:
            logger.debug("query_column:" + str(col))
            query.columns.add(col)

        for col in metadata_columns:
            logger.debug("query_metadata_column:" + str(col))
            query.metadata_columns.add(col)

        if metadata_filter:
            logger.debug("query_metadata_filter:" + str(metadata_filter))
            query.metadata_filter = metadata_filter

        if label_column:
            query.label_column = label_column

        if combine_labels:
            query.combine_labels = combine_labels

        if capture_configurations:
            query.capture_configurations = capture_configurations

        if isinstance(segmenter, str):
            segmenters = self.list_segmenters()
            segmenter_id = segmenters[segmenters["name"] == segmenter].id
            if segmenter_id.shape[0] != 1:
                raise Exception("Segmenter {} not found".format(segmenter))
            query.segmenter = int(segmenter_id[0])
        else:
            query.segmenter = segmenter

        if new:
            query.insert(renderer=renderer)
        else:
            query.update(renderer=renderer)

        return query

    def get_query(self, name):
        if self.project is None:
            print("Project must be set first")
            return

        return self.project.queries.get_query_by_name(name)

    def upload_data_file(self, name, path, force=False, is_features=False):
        """Upload data source from local CSV file"""
        logger.debug("set_feature_file:" + name + ":" + path)
        print('Uploading file "{}" to SensiML Cloud.'.format(name))
        if name[-4:] != ".csv":
            name = "{}.csv".format(name)

        feature_file = self._project._feature_files.get_by_name(name)
        if feature_file is None:
            new = True
            feature_file = self._project.featurefiles.new_featurefile()
        else:
            new = False
            if not force:
                raise FeatureFileExistsException()

        feature_file.filename = name
        feature_file.is_features = is_features
        feature_file.path = path
        if new:
            return feature_file.insert()
        else:
            return feature_file.update()

    def upload_dataframe(self, name, dataframe, force=False, is_features=False):
        """Set data source from a pandas DataFrame."""
        logger.debug("set_data:" + name)

        with NamedTemporaryFile(delete=False) as temp:
            dataframe.to_csv(temp.name, index=False)
            logger.debug("set_dataframe:" + name + ":" + temp.name)
            result = self.upload_data_file(
                name, temp.name, force=force, is_features=is_features
            )

        os.remove(temp.name)

        return result

    def clear_session_cache(self):
        for _, _, filenames in os.walk(config_dir):
            for filename in filenames:
                if re.match(r"_token.json$", filename):
                    os.unlink(filename)

    def get_feature_statistics_results(self, query_name):
        query = self._project.queries.get_query_by_name(query_name)

        for i in range(100):
            results = query.get_feature_statistics()

            if results.get("results", None):
                return pd.DataFrame(results["results"])
            sleep(5)

        print("Not able to reach the results in the expected time. ")
        return results

    def upload_project(self, name, dclprojpath, asynchronous=True):
        if name in self.list_projects()["Name"].values:
            print("Project with this name already exists.")
            return

        self.project = name
        capture_metadata = None
        basedir = os.path.dirname(dclprojpath)
        capture_dir = os.path.join(basedir, "data")
        segment_map = {}

        with open(dclprojpath, "r") as f:
            dclproj = json.load(f)

        plugin_config = dclproj.get("ProjectCapturePluginConfig", None)
        if plugin_config is not None:
            self.project.plugin_config = plugin_config
            print("Capture Config Found, Updating...")
            self.project.update()

        for segmenter in dclproj.get("ProjectSegmenters"):

            if segmenter.get("parameters", None):
                params = json.loads(segmenter.get("parameters"))
                call = self.functions.create_function_call(params["name"])
                for k, v in params["inputs"].items():
                    setattr(call, k, v)
            else:
                call = None

            new_segmenter = self.project.add_segmenter(
                segmenter.get("name"),
                call,
                preprocess=segmenter.get("preprocess"),
                custom=segmenter.get("custom"),
            )

            segment_map[segmenter["name"]] = new_segmenter["id"]
            segment_map[segmenter["id"]] = new_segmenter["id"]

        capture_files = os.listdir(capture_dir)
        for capture in capture_files:
            print("Uploading Capture data, metadata and lables for {}".format(capture))
            try:

                capture_metadata = {}
                file_metadata = []
                if os.path.exists(
                    os.path.join(
                        basedir, "metadata", os.path.basename(capture)[:-3] + "dcl"
                    )
                ):
                    with open(
                        os.path.join(
                            basedir, "metadata", os.path.basename(capture)[:-3] + "dcl"
                        ),
                        "r",
                    ) as f:
                        capture_metadata = json.load(f)
                        file_metadata = capture_metadata.pop("FileMetadata")

                capture_obj = self.project.captures.create_capture(
                    os.path.basename(capture),
                    os.path.join(capture_dir, capture),
                    capture_info=capture_metadata,
                    asynchronous=asynchronous,
                )
                for metadata in file_metadata:
                    print("metadata", metadata["name"], metadata["value"])
                    metadata_obj = capture_obj.metadataset.new_metadata()
                    metadata_obj.value = metadata["value"]
                    metadata_obj.name = metadata["name"]
                    metadata_obj.insert()

                for segment_dir in os.listdir(os.path.join(basedir, "segments")):
                    if not os.path.exists(
                        os.path.join(
                            basedir,
                            "segments",
                            segment_dir,
                            os.path.basename(capture)[:-3] + "sdcl",
                        )
                    ):
                        continue

                    with open(
                        os.path.join(
                            basedir,
                            "segments",
                            segment_dir,
                            os.path.basename(capture)[:-3] + "sdcl",
                        ),
                        "r",
                    ) as f:
                        segment_labels = json.load(f).pop("DetectedSegments")

                    for segment in segment_labels:
                        for label in segment["LabelData"]:
                            print("label", label["name"], label["value"])
                            label_obj = capture_obj.labelset.new_label()
                            label_obj.sample_start = label[
                                "capture_sample_sequence_start"
                            ]
                            label_obj.sample_end = label["capture_sample_sequence_end"]
                            label_obj.value = label["value"]
                            label_obj.name = label["name"]
                            label_obj.segmenter = segment_map[label["segmenter"]]
                            label_obj.insert()

            except CaptureExistsError:
                print("{} already uploaded to this project.".format(capture))



from ipywidgets import (
    Layout,
    Button,
    VBox,
    HBox,
    Dropdown,
    Checkbox,
    Label,
    widgets,
    HTML,
    GridspecLayout,
)
from sensiml.widgets.base_widget import BaseWidget
import qgrid
from pandas import DataFrame


class FeatureGeneratorWidget(BaseWidget):
    def _refresh(self):
        if not self._dsk:
            raise Exception("Must have dsk.")

        if not self._dsk.project:
            raise Exception("Must have active project.")

        if not self._dsk.pipeline:
            raise Exception("Must have active pipeline.")

        fill_widgets_grid(self._dsk, self._w_grid)

    def hide(self):
        self._w_grid.layout.height = "0px"

    def show(self):
        self._w_grid.layout.height = "auto"

    def _clear(self):
        for index, subtype in enumerate(self._w_subtype):
            subtype.children[0].value = False

    def get_subtype_list(self):
        subtype_list = get_active_subtypes(self._w_grid)
        return subtype_list

    def get_fature_generator(self):

        subtype_list = get_active_subtypes(self._w_grid)

        feature_generators = self._dsk.pipeline.add_feature_generator(
            [{"subtype_call": str(x)} for x in subtype_list],
            function_defaults={"columns": self._dsk.pipeline.data_columns},
            return_generator_set=True,
        )._to_dict()

        feature_generators["outputs"] = [
            "temp.generator_set0",
            "temp.features.generator_set0",
        ]

        return feature_generators

    def create_widget(self):

        self._w_grid = build_widgets_grid()

        return self._w_grid


def get_subtype_defaults(dsk, subtype):
    fgs = dsk.list_functions(
        functype="Feature Generator", subtype=subtype, kp_functions=True
    ).df.NAME
    params = {}
    for fg in fgs:
        for param in dsk.functions._function_list[fg].input_contract:
            if param["name"] in ["input_data", "columns", "group_columns"]:
                continue
            params[param["name"]] = param["type"]

    return params


def build_widgets_grid(width=3, height=6):
    grid = GridspecLayout(height, width)
    for row in range(height):
        for col in range(width):
            grid[row, col] = widgets.Checkbox(
                value=False,
                description="{}{}".format(row, col),
                layout=Layout(visibility="hidden"),
            )

    return grid


def fill_widgets_grid(dsk, grid):

    fg_subtypes = dsk.list_functions(functype="Feature Generator").df.SUBTYPE.unique()

    index = 0
    for row in range(grid.n_rows):
        for col in range(grid.n_columns):
            if index >= len(fg_subtypes):
                continue
            grid[row, col].value = False
            grid[row, col].description = fg_subtypes[index]
            grid[row, col].layout.visibility = "visible"
            index += 1

    return grid


def get_active_subtypes(grid):
    active = []
    for row in range(grid.n_rows):
        for col in range(grid.n_columns):
            if grid[row, col].value:
                active.append(grid[row, col].description)

    return active


def build_pipeline(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {"subtype_call": subtype}
        subtype_list.append(tmp_dict)

    dsk.pipeline.add_feature_generator(
        subtype_list, function_defaults={"columns": dsk.pipeline.data_columns}
    )


def build_subtype_list(dsk, active_subtypes):
    subtype_list = []
    for subtype in active_subtypes:
        tmp_dict = {
            "subtype_call": subtype,
            "params": get_subtype_defaults(dsk, subtype),
        }
        subtype_list.append(tmp_dict)

    return subtype_list



class DataSegment(object):
    def __init__(self, data, segment_index, label=None, uuid=None, capture=None):
        self._metadata = ["label", "segment_index", "uuid", "capture"]
        self._label = label
        self._segment_index = segment_index
        self._uuid = uuid
        self._original_index = data.index
        self._capture = capture
        self._data = data.reset_index(drop=True)

    def plot(self, **kwargs):
        self._data.plot(title=self.__str__(), **kwargs)

    @property
    def capture(self):
        return self._capture

    @property
    def segment_index(self):
        return self._segment_index

    @property
    def label(self):
        return self._label

    @property
    def uuid(self):
        return self._uuid

    @property
    def data(self):
        return self._data

    @property
    def metadata(self):
        return {metadata: getattr(self, metadata) for metadata in self._metadata}

    def __str__(self):
        return " ".join(["{}: {}, ".format(k, v) for k, v in self.metadata.items()])


class DataSegments(dict):
    def plot(self, **kwargs):
        import matplotlib as plt

        plt.rcParams.update({"figure.max_open_warning": 0})
        for _, segment in self.items():
            segment.plot(**kwargs)


def to_datasegments(data, metdata_columns, label_column):
    """Converts a dataframe into a data segments object"""

    group_columns = metdata_columns + [label_column]
    g = data.groupby(group_columns)
    ds = DataSegments()

    data_columns = [x for x in data.columns if x not in group_columns]

    for key in g.groups.keys():

        metadata = {}
        for index, value in enumerate(group_columns):
            metadata[value] = key[index]

        tmp_df = g.get_group(key)[data_columns]

        ds[tmp_df["uuid"]] = DataSegment(tmp_df, metadata, metadata[label_column])

    return ds



import os
import datetime
import sensiml.base.utility as utility
from sensiml.datamanager.base import Base


class FeatureFile(Base):
    """Base class for a featurefile object."""

    _fields = ["uuid", "name", "created_at", "last_modified", "is_features"]

    def __init__(
        self, connection, project, name="", path="", is_features=True, uuid=None
    ):
        self._connection = connection
        self._project = project
        self.uuid = uuid
        self._created_at = None
        self._is_features = is_features
        self.name = name
        self.path = path

    # Maintain campatibility with old filename attr
    @property
    def filename(self):
        """The name of the file as stored on the server

        Note:
            Filename must contain a .csv or .arff extension
        """
        return self.name

    @filename.setter
    def filename(self, value):
        self.name = value

    @property
    def is_features(self):
        """If this is a DataFile or FeatureFile"""
        return self._is_features

    @is_features.setter
    def is_features(self, value):
        self._is_features = value

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        self._created_at = datetime.datetime.strptime(
            value[:-6], "%Y-%m-%dT%H:%M:%S.%f"
        )

    def insert(self):
        """Calls the REST API to insert a new featurefile."""
        url = "project/{0}/featurefile/".format(self._project.uuid)
        featurefile_info = {"name": self.name, "is_features": self.is_features}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(url, self.path, featurefile_info, "rb")
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data["uuid"]

        return response

    def update(self):
        """Calls the REST API to update the featurefile's properties on the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        featurefile_info = {"name": self.name, "is_features": self.is_features}

        if not os.path.exists(self.path):
            raise OSError(
                "Cannot update featurefile because the system cannot find the file path: '{0}'. "
                "Please update the file's path attribute.".format(self.path)
            )

        response = self._connection.file_request(
            url, self.path, featurefile_info, "rb", method="put"
        )
        response_data, err = utility.check_server_response(response)

        return response

    def delete(self):
        """Calls the REST API and deletes the featurefile from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("delete", url)
        response_data, err = utility.check_server_response(response)

        return response

    def refresh(self):
        """Calls the REST API and populate the featurefile's properties from the server."""
        url = "project/{0}/featurefile/{1}/".format(self._project.uuid, self.uuid)
        response = self._connection.request("get", url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.name = response_data["name"]
            self.is_features = response_data["is_features"]

        return response

    def download(self):
        """Calls the REST API and retrieves the featurefile's binary data.

        Returns:
            featurefile contents
        """

        url = "project/{0}/featurefile/{1}/data/".format(self._project.uuid, self.uuid)

        response = self._connection.request("get", url)

        return response

    def download_json(self):
        """Calls the REST API and retrieves the featurefile's json data.

        Returns:
            featurefile contents as json
        """

        url = "project/{0}/featurefile/{1}/json/".format(self._project.uuid, self.uuid)

        response = self._connection.request("get", url)

        return response



from sensiml.client import SensiML

try:
    from sensiml.widgets import DashBoard

    __all__ = ["SensiML", "DashBoard"]
except ImportError:
    # dev version fails here so lets import a dummy
    __all__ = ["SensiML"]


try:
    from IPython.core.display import HTML

    display(HTML("<style>.container { width:90% !important; }</style>"))
except:
    pass

name = "sensiml"
__version__ = "2021.2.1"



from sensiml.datamanager.project import Project
import sensiml.base.utility as utility


class ProjectExistsError(Exception):
    """Base class for a project exists error"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Projects:
    """Base class for a collection of projects."""

    def __init__(self, connection):
        self._connection = connection

    def create_project(self, name):
        """Creates a project using the name property.

        Args:
            name (str): name of the new project

        Returns:
            project

        Raises:
            ProjectExistsError, if the project already exists on the server
        """
        if self.get_project_by_name(name) is not None:
            raise ProjectExistsError("project {0} already exists.".format(name))
        else:
            project = self.new_project()
            project.name = name
            project.insert()
            return project

    def get_or_create_project(self, name):
        """Calls the REST API and gets the project by name, if it doesn't exist insert a new project

        Args:
            name (str): name of the project

        Returns:
            project object
        """
        project = self.get_project_by_name(name)

        if project is None:
            print("Project {} does not exist, creating a new project.".format(name))
            project = self.create_project(name)

        return project

    def get_project_by_name(self, name):
        """gets a project from the server using its name property

        Args:
            name (str): name of the project

        Returns:
            project or None if project does not exist
        """
        project_list = self.get_projects()
        for project in project_list:
            if project.name == name:
                if project.schema and not project.query_optimized:
                    if len(project._captures.get_captures()):
                        project.query_optimize()
                return project

        return None

    def __getitem__(self, key):
        return self.get_project_by_name(key)

    def new_project(self, dict={}):
        """Creates a new project.

        Args:
            dict (dict): dictionary containing the attributes of the new project

        Returns:
            project
        """
        project = Project(self._connection)
        if len(dict) is not 0:
            project.initialize_from_dict(dict)
        return project

    def get_projects(self):
        """Gets all projects from the server as project objects.

        Returns:
            list[project]
        """
        url = "project/"
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        projects = []
        if err is False:
            # Populate each project from the server if there was no error.
            for project_params in response_data:
                projects.append(self.new_project(project_params))

        return projects

    def get_projects(self):
        """Gets all projects from the server as project objects.

        Returns:
            list[project]
        """
        url = "project/"
        response = self._connection.request("get", url)
        try:
            response_data, err = utility.check_server_response(response)
        except ValueError:
            print(response)

        projects = []
        if err is False:
            # Populate each project from the server if there was no error.
            for project_params in response_data:
                projects.append(self.new_project(project_params))

        return projects

    def build_project_dict(self):
        """Populates the function_list property from the server."""
        project_dict = {}

        response = self.get_projects()
        for project in response:
            project_dict[project.name] = project

        return project_dict
